@@ -382,7 +382,7 @@ struct connman_network *connman_network_create(const char *identifier,
 	network->type       = type;
 	network->identifier = ident;
 
-	network_list = g_slist_append(network_list, network);
+	network_list = g_slist_prepend(network_list, network);
 
 	return network;
 }
@@ -464,36 +464,19 @@ void connman_network_set_index(struct connman_network *network, int index)
 	struct connman_service *service;
 	struct connman_ipconfig *ipconfig;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		goto done;
 
 	ipconfig = __connman_service_get_ip4config(service);
+	if (ipconfig == NULL)
+		goto done;
+
+	/* If index changed, the index of ipconfig must be reset. */
+	__connman_ipconfig_set_index(ipconfig, index);
 
 	DBG("index %d service %p ip4config %p", network->index,
 		service, ipconfig);
-
-	if (network->index < 0 && ipconfig == NULL) {
-
-		ipconfig = __connman_service_get_ip4config(service);
-		if (ipconfig == NULL)
-			/*
-			 * This is needed for plugins that havent set their
-			 * ipconfig layer yet, due to not being able to get
-			 * a network index prior to creating a service.
-			 */
-			__connman_service_create_ip4config(service, index);
-		else
-			__connman_ipconfig_set_index(ipconfig, index);
-
-	} else {
-		/* If index changed, the index of ipconfig must be reset. */
-		if (ipconfig == NULL)
-			goto done;
-
-		__connman_ipconfig_set_index(ipconfig, index);
-	}
-
 done:
 	network->index = index;
 }
@@ -647,7 +630,7 @@ int connman_network_set_associating(struct connman_network *network,
 	if (associating == TRUE) {
 		struct connman_service *service;
 
-		service = __connman_service_lookup_from_network(network);
+		service = connman_service_lookup_from_network(network);
 		__connman_service_ipconfig_indicate_state(service,
 					CONNMAN_SERVICE_STATE_ASSOCIATION,
 					CONNMAN_IPCONFIG_TYPE_IPV4);
@@ -663,7 +646,7 @@ static void set_associate_error(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_error(service,
 					CONNMAN_SERVICE_ERROR_CONNECT_FAILED);
@@ -673,7 +656,7 @@ static void set_configure_error(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_error(service,
 					CONNMAN_SERVICE_ERROR_CONNECT_FAILED);
@@ -683,7 +666,7 @@ static void set_invalid_key_error(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_error(service,
 					CONNMAN_SERVICE_ERROR_INVALID_KEY);
@@ -693,7 +676,7 @@ static void set_connect_error(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_error(service,
 					CONNMAN_SERVICE_ERROR_CONNECT_FAILED);
@@ -705,7 +688,7 @@ void connman_network_set_ipv4_method(struct connman_network *network,
 	struct connman_service *service;
 	struct connman_ipconfig *ipconfig;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return;
 
@@ -722,7 +705,7 @@ void connman_network_set_ipv6_method(struct connman_network *network,
 	struct connman_service *service;
 	struct connman_ipconfig *ipconfig;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return;
 
@@ -736,7 +719,7 @@ void connman_network_set_ipv6_method(struct connman_network *network,
 void connman_network_set_error(struct connman_network *network,
 					enum connman_network_error error)
 {
-	DBG("nework %p, error %d", network, error);
+	DBG("network %p error %d", network, error);
 
 	network->connecting = FALSE;
 	network->associating = FALSE;
@@ -773,7 +756,7 @@ void connman_network_clear_error(struct connman_network *network)
 	if (network->connecting == TRUE || network->associating == TRUE)
 		return;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	__connman_service_clear_error(service);
 }
 
@@ -791,7 +774,7 @@ static void set_configuration(struct connman_network *network,
 
 	connman_device_set_disconnected(network->device, FALSE);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	__connman_service_ipconfig_indicate_state(service,
 					CONNMAN_SERVICE_STATE_CONFIGURATION,
 					type);
@@ -803,7 +786,7 @@ static void dhcp_success(struct connman_network *network)
 	struct connman_ipconfig *ipconfig_ipv4;
 	int err;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		goto err;
 
@@ -831,7 +814,7 @@ static void dhcp_failure(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return;
 
@@ -859,7 +842,7 @@ static int set_connected_fixed(struct connman_network *network)
 
 	DBG("");
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	ipconfig_ipv4 = __connman_service_get_ip4config(service);
 
@@ -894,7 +877,7 @@ static void set_connected_manual(struct connman_network *network)
 
 	DBG("network %p", network);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	ipconfig = __connman_service_get_ip4config(service);
 
@@ -948,13 +931,15 @@ static int manual_ipv6_set(struct connman_network *network,
 
 	DBG("network %p ipv6 %p", network, ipconfig_ipv6);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
 	if (__connman_ipconfig_get_local(ipconfig_ipv6) == NULL)
 		__connman_service_read_ip6config(service);
 
+	__connman_ipconfig_enable_ipv6(ipconfig_ipv6);
+
 	err = __connman_ipconfig_address_add(ipconfig_ipv6);
 	if (err < 0) {
 		connman_network_set_error(network,
@@ -1011,7 +996,7 @@ static gboolean dhcpv6_set_addresses(struct connman_network *network)
 	struct connman_ipconfig *ipconfig_ipv6;
 	int err = -EINVAL;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		goto err;
 
@@ -1203,7 +1188,7 @@ static void autoconf_ipv6_set(struct connman_network *network)
 
 	network->connecting = FALSE;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return;
 
@@ -1232,7 +1217,7 @@ static void set_connected(struct connman_network *network)
 
 	network->connected = TRUE;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	ipconfig_ipv4 = __connman_service_get_ip4config(service);
 	ipconfig_ipv6 = __connman_service_get_ip6config(service);
@@ -1304,7 +1289,7 @@ static void set_disconnected(struct connman_network *network)
 
 	network->connected = FALSE;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	ipconfig_ipv4 = __connman_service_get_ip4config(service);
 	ipconfig_ipv6 = __connman_service_get_ip6config(service);
@@ -1468,7 +1453,7 @@ void connman_network_clear_hidden(void *user_data)
 	 * error to the caller telling that we could not find
 	 * any network that we could connect to.
 	 */
-	__connman_service_reply_dbus_pending(user_data, EIO);
+	__connman_service_reply_dbus_pending(user_data, EIO, NULL);
 }
 
 int connman_network_connect_hidden(struct connman_network *network,
@@ -1477,7 +1462,7 @@ int connman_network_connect_hidden(struct connman_network *network,
 	int err = 0;
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	DBG("network %p service %p user_data %p", network, service, user_data);
 
@@ -1592,7 +1577,7 @@ static int manual_ipv4_set(struct connman_network *network,
 	struct connman_service *service;
 	int err;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
@@ -1613,7 +1598,7 @@ int __connman_network_clear_ipconfig(struct connman_network *network,
 	enum connman_ipconfig_method method;
 	enum connman_ipconfig_type type;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
@@ -1709,7 +1694,7 @@ int connman_network_set_ipaddress(struct connman_network *network,
 
 	DBG("network %p", network);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
@@ -1735,7 +1720,7 @@ int connman_network_set_nameservers(struct connman_network *network,
 
 	DBG("network %p nameservers %s", network, nameservers);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
@@ -1763,7 +1748,7 @@ int connman_network_set_domain(struct connman_network *network,
 
 	DBG("network %p domain %s", network, domain);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
