@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -165,7 +165,7 @@ int __connman_rfkill_block(enum connman_service_type type, connman_bool_t block)
 	uint8_t rfkill_type;
 	struct rfkill_event event;
 	ssize_t len;
-	int fd;
+	int fd, err;
 
 	DBG("type %d block %d", type, block);
 
@@ -183,12 +183,15 @@ int __connman_rfkill_block(enum connman_service_type type, connman_bool_t block)
 	event.soft = block;
 
 	len = write(fd, &event, sizeof(event));
-	if (len < 0)
+	if (len < 0) {
 		connman_error("Failed to change RFKILL state");
+		err = len;
+	} else
+		err = 0;
 
 	close(fd);
 
-	return 0;
+	return err;
 }
 
 int __connman_rfkill_init(void)
