@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -71,30 +71,20 @@ void util_session_cleanup(struct test_session *session);
 
 typedef void (* notify_cb) (struct test_session *session);
 
-enum connman_session_roaming_policy {
-	CONNMAN_SESSION_ROAMING_POLICY_UNKNOWN		= 0,
-	CONNMAN_SESSION_ROAMING_POLICY_DEFAULT		= 1,
-	CONNMAN_SESSION_ROAMING_POLICY_ALWAYS		= 2,
-	CONNMAN_SESSION_ROAMING_POLICY_FORBIDDEN	= 3,
-	CONNMAN_SESSION_ROAMING_POLICY_NATIONAL		= 4,
-	CONNMAN_SESSION_ROAMING_POLICY_INTERNATIONAL	= 5,
+enum connman_session_state {
+	CONNMAN_SESSION_STATE_DISCONNECTED   = 0,
+	CONNMAN_SESSION_STATE_CONNECTED      = 1,
+	CONNMAN_SESSION_STATE_ONLINE         = 2,
 };
 
 struct test_session_info {
-	char *bearer;
-	connman_bool_t online;
+	enum connman_session_state state;
 	char *name;
+	char *bearer;
+	char *interface;
+	enum connman_session_type type;
 	/* ipv4, ipv6 dicts */
 	GSList *allowed_bearers;
-	connman_bool_t priority;
-	connman_bool_t avoid_handover;
-	connman_bool_t stay_connected;
-	unsigned int periodic_connect;
-	unsigned int idle_timeout;
-	connman_bool_t ecall;
-	enum connman_session_roaming_policy roaming_policy;
-	char *interface;
-	unsigned int marker;
 };
 
 struct test_session {
